;  3     Capture Input ( TSOP1736)
;                                       (C)  .. 15.01.2004 16:45

;          Capture  Input    
;      .      start
; -,         .         
;   +-1/PR %.      ,    
;               , 
;   ,  ,   .     
;     (       
;     ),   .
;    - MSB .

.include "2313def.inc"

;  
.equ PinE      = 3
.equ PinRW     = 2
.equ PinRS     = 5
.equ LED       = 2

.equ    PR = 1500                       ;   (*100ns)

.def    rSpaceTableWorkEntrieL  = r8
.def    rSpaceTableWorkEntrieH  = r9
.def    rLEDTimerCount  = r10           ;    InterruptTimer0
.def    rUartTxState    = r11
.def    rIRRecState     = r25
.def    rFlags          = r24
;  Init    !
.equ        Initl = 0                   ;   UART'
.equ        Inith = 1
.equ        IRReceiveComplite = 2       ; == 1 -  
.equ        TXAck1 = 'I'
.equ        TXAck2 = 'R'

            rjmp    MainStart
            reti
            reti
            rjmp    InterruptInputCapture
            reti
            rjmp    InterruptTimer1Ovr
            rjmp    InterruptTimer0
            rjmp    RXComplite
            rjmp    UDREmpty
            reti


MainStart:  clr     r17                         ; wait full power
            clr     r16
            dec     r16
            brne    PC-1
            dec     r17
            brne    PC-4

            ldi     r16,low(RAMEND)             ; set stack
            out     SPL,r16
            ldi     r16,0b00111100
            out     DDRD,r16                    ; setting port D
            sbi     PORTD,0                     ; pull-up on RX-pin
            sbi     PORTD,6
            ldi     r16,0b11111100              ; setting port B
            out     DDRB,r16
            ldi     r16,(1<<ICNC1)|(1<<CS10)    ; Input Capture,
            out     TCCR1B,r16                  ; Noise Canc., clk - CK
            ldi     r16,(1<<TICIE)|(1<<TOIE1)|(1<<TOIE0)
            out     TIMSK,r16
            ldi     r16,(1<<ACD)                ; Analog Comparator - desabled
            out     ACSR,r16
            clr     r16
            mov     rLEDTimerCount,r16
            out     TCNT0,r16


            clr     rIRRecState
            clr     rFlags
            ldi     r16,(1<<SE)
            out     MCUCR,r16
            ldi     r16,(1<<RXCIE)|(1<<RXEN)|(1<<TXEN)
            out     UCR,r16
            ldi     r16,64                      ; 9600 for 10 MHz
            out     UBRR,r16
            ser     r16
            mov     rUartTxState,r16
            sei

stop:       sleep                               ;   
            sbrs    rFlags,IRReceiveComplite
            rjmp    stop
            mov     r16,rFlags                  ; ,  -
            andi    r16,3                       ;    UART'
            cpi     r16,3
            brne    stop
            rcall   CalcCRC                     ;  CRC  
            ldi     YL,low(OutputBuffer)        ;  ...
            ldi     YH,high(OutputBuffer)
            ld      r16,Y+
            out     UDR,r16
            clr     rUartTxState
            inc     rUartTxState
            in      r16,UCR
            sbr     r16,(1<<UDRIE)              ; .   UDREmpty
            out     UCR,r16

            cbr     rFlags,(1<<IRReceiveComplite)
            rjmp    stop



;   
CalcCRC:    ldi     ZL,low(OutputBuffer)
            ldi     ZH,high(OutputBuffer)
            ldi     r16,RawBufferSize
            ldi     r18,$0AA                    ;   CRC
            ldi     r20,$050                    ; CRC 
crc1:       ld      r17,Z+
            eor     r18,r17
            ldi     r19,8
crc2:       clc
            rol     r18
            brcc    PC+2
            eor     r18,r20
            dec     r19
            brne    crc2
            dec     r16
            brne    crc1
            sts     OutputBufferCRC,r18
            ret

;//////// Disable LED /////////////////////
; Off LED after 1e-7 * 1024 * 256 * 2 = 52 ms
InterruptTimer0:
            push    r16
            in      r16,SREG
            push    r16
            mov     r16,rLEDTimerCount
            inc     r16
            cpi     r16,2
            brcs    IT1
            clr     r16                         ; rLEDTimerCount = 0
            cbi     PORTD,LED                   ; LED off
            out     TCCR0,r16                   ; Timer0 stop
IT1:        mov     rLEDTimerCount,r16
            pop     r16
            out     SREG,r16
            pop     r16
            reti

;//////// Work with UART //////////////////
RXComplite:
            push    r16
            in      r16,SREG
            push    r16
            sbis    USR,FE                      ; check Frame Error
            rjmp    RXInit0
            in      r16,UDR
            rjmp    RXExit
RXInit0:
            mov     r16,rFlags
            andi    r16,3
            breq    RXInit1
            dec     r16
            breq    RXInit2
            cbr     rFlags,(1<<Initl)|(1<<Inith)

RXInit1:    in      r16,UDR
            cpi     r16,'H'
            brne    RXExit
            sbr     rFlags,1<<Initl
            rjmp    RXExit

RXInit2:    in      r16,UDR
            cbr     rFlags,1<<Initl
            cpi     r16,'i'
            brne    RXExit
            sbr     rFlags,1<<Inith             ; Init = 2
            ldi     r16,TXAck1
            out     UDR,r16
            in      r16,UCR
            sbr     r16,(1<<UDRIE)              ; on UDREmpty int.
            out     UCR,r16

RxExit:     pop     r16
            out     SREG,r16
            pop     r16
            reti


UDREmpty:   push    r16
            in      r16,SREG
            push    r16
            push    r17
            mov     r16,rFlags
            andi    r16,3
            cpi     r16,2
            breq    UDREmptyInit
            ldi     r16,OutputBufferSize
            cp      rUartTxState,r16
            brcc    UDREmptyOff
            inc     rUartTxState
            ld      r16,Y+
            out     UDR,r16
            rjmp    UDREmptyExit
UDREmptyInit:
            ldi     r16,TXAck2
            out     UDR,r16
            sbr     rFlags,1<<Initl
UDREmptyOff:
            in      r16,UCR
            cbr     r16,(1<<UDRIE)              ; off UDREmpty int.
            out     UCR,r16
UDREmptyExit:
            pop     r17
            pop     r16
            out     SREG,r16
            pop     r16
            reti

;////////   - ////////////
InterruptTimer1Ovr:
            push    r16
            in      r16,SREG
            push    r16

            in      r16,TCCR1B
            cbr     r16,1 << ICES1
            out     TCCR1B,r16
            clr     rIRRecState

            pop     r16
            out     SREG,r16
            pop     r16
            reti


InterruptInputCapture:
            push    r16
            in      r16,SREG
            push    r16
            push    YL
            push    YH
            push    ZL
            push    ZH
            push    r0

            tst     rIRRecState
            breq    ICP1Start
            in      YL,ICR1L
            in      YH,ICR1H
            cpi     rIRRecState,3
            brcs    ICP1Header
            rjmp    ICP1Read

;-----------------
ICP1Start:
            in      r16,TCCR1B
            sbr     r16,1 << ICES1
            out     TCCR1B,r16

ICP1Next:
            in      YL,ICR1L
            in      YH,ICR1H
            in      ZL,TCNT1L
            in      ZH,TCNT1H
            sub     ZL,YL
            sbc     ZH,YH
            out     TCNT1H,ZH                       ; reset TCNT
            out     TCNT1L,ZL
            inc     rIRRecState
            ldi     r16,(1<<TOV1)
            out     TIFR,r16                        ; reset TOV1
            rjmp    ICP1Exit

;-----------------
ICP1Header:
            cpi     rIRRecState,2
            breq    ICP1hFind
            sts     RawBuffer+0,YL
            sts     RawBuffer+1,YH
            in      r16,TCCR1B
            cbr     r16,1 << ICES1
            out     TCCR1B,r16
            inc     rIRRecState
            rjmp    ICP1Exit
ICP1hFind:  sts     RawBuffer+2,YL
            sts     RawBuffer+3,YH
            clr     r8
            ldi     ZL,low(SpaceTimesTable*2 + steHeaderMiddle)
            ldi     ZH,high(SpaceTimesTable*2 + steHeaderMiddle)
ICP1hFindL: lds     YL,RawBuffer+0
            lds     YH,RawBuffer+1
            rcall   InterruptLoadAndCompare
            brcc    ICP1hFindNext_

            adiw    ZL,steHeaderEnd - steHeaderMiddle - 1
            lds     YL,RawBuffer+2
            lds     YH,RawBuffer+3
            rcall   InterruptLoadAndCompare
            brcc    ICP1hFindNext

            sbiw    ZL,steHeaderEnd + 1
            push    r8
            mov     rSpaceTableWorkEntrieL,ZL
            mov     rSpaceTableWorkEntrieH,ZH
            ldi     ZL,low(RawBuffer)
            ldi     ZH,high(RawBuffer)
            ldi     r16,RawBufferSize
            clr     YL
ICP1hClear: st      Z+,YL
            dec     r16
            brne    ICP1hClear
            pop     r16
            sts     RawBuffer+IRTypeOffset,r16
            rjmp    ICP1Next

ICP1hFindNext_:
            adiw    ZL,steHeaderEnd - steHeaderMiddle
ICP1hFindNext:
            adiw    ZL,steEntrieSize - steHeaderEnd - 1
            inc     r8
            ldi     r16,SpaceTimesEntriesCount
            cp      r8,r16
            brcs    ICP1hFindL
            rjmp    ICP1Stop

;-----------------
ICP1Read:
            mov     ZL,rSpaceTableWorkEntrieL
            mov     ZH,rSpaceTableWorkEntrieH
            adiw    ZL,stePulseMiddle
            rcall   InterruptLoadAndCompare
            brcs    ICP1Exit
            clt
            in      YL,ICR1L
            in      YH,ICR1H
            adiw    ZL,stePulse0End - stePulseMiddle - 1
            rcall   InterruptLoadAndCompare
            brcs    ICP1Read0
            set
            in      YL,ICR1L
            in      YH,ICR1H
            adiw    ZL,stePulse1End - stePulse0End - 1
            rcall   InterruptLoadAndCompare
            brcs    ICP1Read1
            rjmp    ICP1Stop
ICP1Read0:  adiw    ZL,stePulse1End - stePulse0End
ICP1Read1:  adiw    ZL,steBitsCount - stePulse1End - 1
            lpm
            mov     r16,rIRRecState
            subi    r16,3
            lsr     r16
            lsr     r16
            lsr     r16
            ldi     ZL,low(RawBuffer)
            ldi     ZH,high(RawBuffer)
            add     ZL,r16
            clr     r16
            adc     ZH,r16
            ld      r16,Z
            ror     r16
            bld     r16,7
            st      Z,r16
            cp      rIRRecState,r0
            brcc    ICP1ReadComplite
            rjmp    ICP1Next

;-----------------
ICP1ReadComplite:
            clr     rIRRecState
            sbi     PORTD,LED
            ldi     r16,(1<<CS02)|(1<<CS00)     ; start Timer0
            out     TCCR0,r16
            sbrc    rFlags,IRReceiveComplite
            rjmp    ICP1Exit

            ldi     ZL,low(RawBuffer)           ;   
            ldi     ZH,high(RawBuffer)          ;  - 
            ldi     r16,RawBufferSize           ;   "value"
ICP1Move:   ld      YL,Z
            std     Z+RawBufferSize,YL
            adiw    ZL,1
            dec     r16
            brne    ICP1Move
            sbr     rFlags,1 << IRReceiveComplite
            rjmp    ICP1Exit

ICP1Stop:
            clr     rIRRecState

ICP1Exit:
            pop     r0
            pop     ZH
            pop     ZL
            pop     YH
            pop     YL
            pop     r16
            out     SREG,r16
            pop     r16
            reti

; r0, r16, Y, Z
; in: Y - number, Z - address
; out: c = 1 -   
InterruptLoadAndCompare:
            lpm
            mov     r16,r0
            adiw    ZL,1
            lpm
            sub     YL,r16
            sbc     YH,r0
            brcc    PC+4
            com     YL
            com     YH
            adiw    YL,1
            subi    YL,low(PR)
            sbci    YH,high(PR)
            ret


;        space- (REC-80)
;   : header_middle, header, pulse_middle, pulse0, pulse1, bits + 2
.equ steEntrieSize   = 12
.equ steHeaderMiddle = 0
.equ steHeaderEnd    = 2
.equ stePulseMiddle  = 4
.equ stePulse0End    = 6
.equ stePulse1End    = 8
.equ steBitsCount    = 10

.equ SpaceTimesEntriesCount = 2

SpaceTimesTable:
          .dw 36000, 53000, 4000, 8000,  16000, 48 + 2      ; (0) Panasoinc, Technics
          .dw 20800, 62400, 5200, 10400, 20800, 16 + 2      ; (1) JVC


.dseg

.equ IRTypeOffset = 6
.equ RawBufferSize = 7
.equ OutputBufferSize = 8

RawBuffer:          .byte 7
OutputBuffer:       .byte 7
OutputBufferCRC:    .byte 1
